// Draconian ARM routines
// Last Update:  May 7, 2014
//              Darrell Spice, Jr.

// To compile, this you'll need to install an ARM compiler.
// for OS X use:
// http://www.gnuarm.com/files.html
//
// this demo was written using the OS X GCC-3.3 toolchain
// http://www.gnuarm.com/binutils-2.14_gcc-3.3.2-c-c++_newlib-1.12.0_gdb-6.0.pkg.tgz
//
//
// for Linux/Windows use:
// http://www.codesourcery.com/sgpp/lite/arm/portal/release1033
//
// You will need to update the "Makefile" if you're using Linux/Windows.  Find the toolchain lines
// and move the # from the aebi toolchain to the elf version.
//  #TOOLCHAIN=arm-elf
//  TOOLCHAIN=arm-none-eabi


// src/custom.h defines the following:
//
// queue[]     - points to the 4K Display Data Bank
//             - treat as RAM
//             - Any data passed to/from the ARM and 6507 must be done via queue[]
//
// flashdata[] - points to the 24K that comprises the six 4K banks.
//             - treat as ROM
#include "src/custom.h"


// types of variables, storage used and range
// const              = constant, which is compiled into the ROM section and cannot be changed during run time.
// signed char        =  8 bit, 1 byte  per value.  Range is           -128 to           127
// unsigned char      =  8 bit, 1 byte  per value.  Range is              0 to           255
// short int          = 16 bit, 2 bytes per value.  Range is        -32,768 to        32,767
// unsigned short int = 16 bit, 2 bytes per value.  Range is              0 to        65,535
// int                = 32 bit, 4 bytes per value.  Range is -2,147,483,648 to 2,147,483,647
// unsigned int       = 32 bit, 4 bytes per value.  Range is              0 to 4,294,967,295
// long int           = same as int
// float              = 32 bit, 4 bytes.  Not Supported#
// double             = 64 bit, 8 bytes.  Not Supported#
// *                  = 32 bit, 4 bytes, pointer (ie: char*, int*)
//
// # - at least not supported with the compiler under OS X.  It might be supported by the Linux/Windows ARM
//     compiler - but support would be implemented via software as the ARM in the Harmony Cartridge does not
//     have an FPU (floating point unit).  As such, performance will most likely be as good as using
//     integer math based routines.
//
// NOTE : Only 448 bytes are allocated for use by variables (the rest of the 512 byte section is used 
//        as the stack).  If you use too many variables, you'll get a compile time error:  "region ram is full"
//        You can use RAM in the Display Data to store values, see defines.h for queue[xxx] defines
//
// NOTE : Compiled code can be significantly smaller if you use INTs for your variables instead of
//        SHORTs or CHARs.


#include "defines.h"
unsigned char * const queue=(unsigned char *)0x40000C00;
unsigned char * const flashdata=(unsigned char *)0x0C00;
unsigned char * const ddrom=(unsigned char *)0x6C00;    // ROM copy of Display Data contents

int * const queue_int=(int *)0x40000C00;

// TIA Player Data Streams
unsigned char * const gP0DataStream             =(unsigned char *)(0x40000C00 + P0_DATASTREAM);
unsigned char * const gP1DataStream             =(unsigned char *)(0x40000C00 + P1_DATASTREAM);
unsigned char * const gMissileBallDataStream    =(unsigned char *)(0x40000C00 + MISSILE_BALL_DATASTREAM);
unsigned char * const gHMP0DataStream           =(unsigned char *)(0x40000C00 + HMP0_DATASTREAM);
unsigned char * const gHMP1DataStream           =(unsigned char *)(0x40000C00 + HMP1_DATASTREAM);
unsigned char * const gColorHMDataStream        =(unsigned char *)(0x40000C00 + COLOR_HM_DATASTREAM);
unsigned char * const gSizeBallColorDataStream  =(unsigned char *)(0x40000C00 + SIZE_BALL_COLOR_DATASTREAM);
unsigned short int * const gJumpDataStream      =(unsigned short int *)(0x40000C00 + JUMP_DATASTREAM);

// Menu Options datastream
unsigned short int * const gMenuOptionsDatastream  =(unsigned short int *)(0x40000C00 + MENU_OPTIONS_DS);
unsigned char * const gMenuColors	=(unsigned char *)(0x40000C00 + MAIN_MENU_COLORS);

// Sprite Data
unsigned char gStationState[8];
unsigned char gSpriteState[MAX_SPRITES];
unsigned char gSpriteImageID[MAX_SPRITES+1];    // +1 is used for Station Pod collision detection
unsigned short int gSpriteX[MAX_SPRITES+1];     // +1 is used for Station Pod collision detection
unsigned short int gSpriteY[MAX_SPRITES+1];     // +1 is used for Station Pod collision detection
unsigned char gSpriteColor[MAX_SPRITES];
unsigned char * const gStationImages =(unsigned char *)(0x40000C00 + STATION_IMAGES);

unsigned char gSpriteFlickerList1[MAX_SPRITES];
unsigned char gSpriteFlickerList2[MAX_SPRITES];
static unsigned char* gNonFlickeredSprites;

unsigned char gShotState[MAX_SHOTS+1];
unsigned char gShotX[MAX_SHOTS+1];
unsigned char gShotY[MAX_SHOTS+1];
unsigned char gShotFlickerList1[MAX_SHOTS];
unsigned char gShotFlickerList2[MAX_SHOTS];
static unsigned char* gNonFlickeredShots;

// Image Data
short int *gImageGraphics           =(short int *)(0x0C00 + IMAGE_GRAPHICS);
short int *gImageShift              =(short int *)(0x0C00 + IMAGE_SHIFT);
unsigned char *gImageHeight         =(unsigned char *)(0x0C00 + IMAGE_HEIGHTS);
unsigned char * const gObjectColors =(unsigned char *)(0x40000C00 + OBJECT_COLORS);

// Score/Radar display data
unsigned char * const gPlayfield    =(unsigned char *)(0x40000C00 + PLAYFIELD_DATA);

int gScore;
int gLevel;

short int gStarfieldXoffset;
short int gStarfieldYoffset;
unsigned char gStarfieldTopStar;
unsigned char gReserveLives;
unsigned char gCondition;

unsigned short int gWindowX;
unsigned short int gWindowY;
int gShipDir;
int gFormationSize;

unsigned char gGunCoolDown;
unsigned char gInitLevel;

#define MAMCR *(unsigned char*)0xE01FC000

// function templates
int Collision(int image1_id, int left1, int top1, int sprite_index);
int ColorConvert(int color);
int DataStream(int stream, int index, int lumaboost);
int DataStreamShot(int stream, int index, int firstcall);
void DeathSequences();
void DisplayShots();
void DisplaySprites();
void DisplayStars();
void FixSizes();
void IncreaseScore(int points);
void InitDatastreams();
void InitDeathSequence(int index, int not_deadly);
void Initialize();
void InitLevel();
void MoveFormation();
void MoveShots();
void my_memcpy(unsigned char* destination, unsigned char* source, int count);   // memory copy routine.
void my_memset(unsigned char* destination, int fill_value, int count);          // memory fill routine
void NewGame();
void OverScan();
void OverScanMenu();
void PodHit(int station, int pod);
void PrepScoreRadar();
void ProcessJoystick();
void ProcessMenuJoystick();
void ProcessSwitches();
unsigned int Random32();
void ShipCollisions();
void ShotCollisions();
void StationShots();
void PrepStationGraphics();
void TriggerSoundEffect(int SoundID);
void UpdateSound();
void VerticalBlank();
void VerticalBlankMenu();

//last two are the stations energy pod
const unsigned char shot_height[] =
{
    1,      // Horizontal Shot
    2,      // Diagonal Shot
    5,      // Vertical Shot
    3,      // Station Shot
    3,      // Horizontal Station Energy Pod 0
    7,      // Horizontal Station Energy Pod 1
    7,      // Vertical Station Energy Pod 0
    13      // Vertical Station Energy Pod 1
};
const unsigned char shot_size[] =
{
    0x20,   // Horizontal Shot
    0x10,   // Diagonal Shot
    0x00,   // Vertical Shot
    0x10,   // Station Shot
    0x30,   // Horizontal Station Energy Pod 0
    0x20,   // Horizontal Station Energy Pod 1
    0x20,   // Vertical Station Energy Pod 0
    0x10    // Vertical Station Energy Pod 1
};

// pods are numbered 0-5 in this arrangement around the station cores:
//  0     0 1
// 5 1   5   2  <- Vertical
// 4 2    4 3
//  3  <---------- Horizontal
const unsigned char pod_offset_x[] =
{
    10, // Horizontal Pod 0
    20, // Horizontal Pod 1
    20, // Horizontal Pod 2
    10, // Horizontal Pod 3
     0, // Horizontal Pod 4
     0, // Horizontal Pod 5
    10, // Horizontal Energy Pod
     6, // Vertical Pod 0
    18, // Vertical Pod 1
    24, // Vertical Pod 2
    18, // Vertical Pod 3
     6, // Vertical Pod 4
     0, // Vertical Pod 5
    12  // Vertical Energy Pod
};

const unsigned char pod_offset_y[] =
{
     0, // Horizontal Pod 0
    13, // Horizontal Pod 1
    37, // Horizontal Pod 2
    50, // Horizontal Pod 3
    37, // Horizontal Pod 4
    13, // Horizontal Pod 5
    28, // Horizontal Energy Pod
     0, // Vertical Pod 0
     0, // Vertical Pod 1
    21, // Vertical Pod 2
    42, // Vertical Pod 3
    42, // Vertical Pod 4
    21, // Vertical Pod 5
    22  // Vertical Energy Pod
};

const unsigned char pod_shape[] =
{
    POD_ID,                     // Horizontal Pod 0
    POD_ID,                     // Horizontal Pod 1
    POD_ID,                     // Horizontal Pod 2
    POD_ID,                     // Horizontal Pod 3
    POD_ID,                     // Horizontal Pod 4
    POD_ID,                     // Horizontal Pod 5
    HORIZONTAL_ENERGY_POD_ID,   // Horizontal Energy Pod
    POD_ID,                     // Vertical Pod 0
    POD_ID,                     // Vertical Pod 1
    POD_ID,                     // Vertical Pod 2
    POD_ID,                     // Vertical Pod 3
    POD_ID,                     // Vertical Pod 4
    POD_ID,                     // Vertical Pod 5
    VERTICAL_ENERGY_POD_ID      // Vertical Energy Pod
};

int main()
{
    MAMCR=1;    // partially disable MAM (cache) to stop fatal bug in ARM chip
    
    switch (SUB)    
    {
        case 0:    Initialize();        break;
        case 1:    OverScan();          break;
        case 2:    VerticalBlank();     break;
        case 3:    OverScanMenu();      break;
        case 4:    VerticalBlankMenu(); break;
        case 5:    InitLevel();         break;
    }
    
    MAMCR=2;    // reenable MAM
    
    return 0;
}

void my_memset(unsigned char* destination, int fill_value, int count)
{
    int i;
    for (i=0;i<count;i++)
        destination[i]=fill_value;
}

void my_memcpy(unsigned char* destination, unsigned char* source, int count)
{
    int i;
    for(i=0;i<count;i++)
        destination[i] = source[i];
}

void Initialize()
{
}


void OverScan()
{
    ProcessSwitches();    // processes SELECT and RESET
    if (MODE != 0)            // if MODE is 0 then we're returning to main menu
    {
        DeathSequences();
        ShotCollisions();   // done after shots were drawn so collisions can be seen
        ShipCollisions();
        ProcessJoystick();
        StationShots();
        MoveShots();
        MoveFormation();
        PrepStationGraphics();
        PrepScoreRadar();
    }
    UpdateSound();
}


void VerticalBlank()
{    
    InitDatastreams();
    if (gInitLevel)
    {
        gJumpDataStream[0] = INIT_LEVEL_KERNEL;
        PrepScoreRadar();
    }
    else
    {
        DisplaySprites();
        DisplayShots();
        FixSizes();
        DisplayStars();
    }
    gP0DataStream[DATASTREAM_SIZE] = gP1DataStream[DATASTREAM_SIZE] = 0;
}


void OverScanMenu()
{
    ProcessSwitches();    // processes SELECT and RESET
    if (MODE == 0)            // if MODE = 1, then we're about to start the game
    {
        ProcessMenuJoystick();
    }
    UpdateSound();
}

void VerticalBlankMenu()
{
    int i;
    int ones;
    int tens;
    int hundreds;
    int thousands;
    
    // copy over all menu graphics and NTSC colors from Display Data ROM as the RAM is repurposed during the game.
    // this also inits the options to their default values, so we just have to change those that are different.
    my_memcpy((unsigned char *)queue + MAIN_MENU_GRAPHICS,
              (unsigned char *)ddrom + MAIN_MENU_GRAPHICS,
              MAIN_MENU_GFX_SIZE);
    
    // Convert colors based on MM_TV_TYPE
    for(i=0;i<MAIN_MENU_CLR_SIZE;i++)
        gMenuColors[i] = ColorConvert(gMenuColors[i]);
    
    // Convert object colors based on MM_TV_TYPE
    // -2 so the RadarColors also get updated
    for(i=-2;i<OBJECT_COLOR_SIZE;i++)
        queue[OBJECT_COLORS + i] = ColorConvert(ddrom[OBJECT_COLORS+i]);
    
    // start create level number display
    thousands = 0;
    hundreds = 0;
    tens = 0;
    ones = MM_INITIAL_LEVEL+1; // show level as +1 as humans prefer "level 1" as the first level and not "level 0"
    
    while (ones >= 1000)
    {
        thousands++;
        ones -= 1000;
    }
    while (ones >= 100)
    {
        hundreds++;
        ones -= 100;
    }
    while (ones >= 10)
    {
        tens++;
        ones -= 10;
    }
    if (thousands == 0)
    {
        thousands = 10;     // leading space
        if (hundreds == 0)
        {
            hundreds = 10;  // leading space
            if (tens == 0)
                tens = 10;  // leading space
        }
    }
    
    for(i=0;i<8;i++)
    {
        LEVEL_DIGITS_A[i] = (MENU_DIGITS[i + thousands*8] << 4) + MENU_DIGITS[i + hundreds*8];
        LEVEL_DIGITS_B[i] = (MENU_DIGITS[i + tens*8] << 4)      + MENU_DIGITS[i + ones*8];
    }
    // end create level number display
    
    switch (MM_TV_TYPE)
    {   // TV Type: [0]=NTSC, 1=PAL, 2=SECAM
        case 1: gMenuOptionsDatastream[MENU_TV_TYPE_ID*10 + 0] = OPTION_PAL_B;
            gMenuOptionsDatastream[MENU_TV_TYPE_ID*10 + 1] = OPTION_PAL_A;
            break;
        case 2: gMenuOptionsDatastream[MENU_TV_TYPE_ID*10 + 0] = OPTION_SECAM_C;
            gMenuOptionsDatastream[MENU_TV_TYPE_ID*10 + 1] = OPTION_SECAM_B;
            gMenuOptionsDatastream[MENU_TV_TYPE_ID*10 + 2] = OPTION_SECAM_A;
            break;
    }
    
    // flash selected option
    for(i=0;i<16;i+=2)
    {
        
        // this resulted in selected item not being visible on SECAM Atari's until SECAM had been selected...
        //        if (MM_TV_TYPE == 2)
        //            queue[MENU_OPTION_SELECTED + i+1] = FRAME;
        //        else
        //            queue[MENU_OPTION_SELECTED + i+1] |= (FRAME << 4);
        
        queue[MENU_SELECTED_COLOR + i+1] = FRAME;  // visible on all systems
        queue[MENU_SELECTED_COLOR + i+0] = queue[gMenuOptionsDatastream[MM_SELECTED_OPTION*10 + 6] + i];
    }
    gMenuOptionsDatastream[MM_SELECTED_OPTION*10 + 6] =  MENU_SELECTED_COLOR;
}

void ProcessSwitches()
{
    FRAME++;
    Random32();                                                            // mix up the random numbers
    if (!(SWCHB & 2))               MODE = 0;     // show the main menu
    else if (!(SWCHB & 1))          NewGame();        // start new game
}



void NewGame()
{
    MODE = 1;
    gScore = 0;
    gReserveLives = 3;
    gCondition = 0;
    gGunCoolDown = 0;
    
    gLevel = MM_INITIAL_LEVEL;
    
//    InitLevel();
    gInitLevel = 1;
}


// if possible try to position stations so they
// are centered in the area covered by a single
// radar pixel.
//
//      rx = 0-7 (radar pixel X range)
//      ry = 015 (radar pixel Y range)
//
//  HORIZONTAL STATIONS
//      x = rx*64 + 20
//      y = ry*64 + 8
//
//  VERTICAL STATIONS
//      x = rx*64 + 20
//      y = ry*64 + 12
//
// level 1 positions have been updated

const unsigned short int level1[] =
{
    256, 512,   // ship starting X, Y location
    8,          // station count
    HS + 20, 8,    // station 1 X, Y location
    VS + 84, 72,   // station 2 X, Y location
    HS + 148, 136,   // station 3 X, Y location
    VS + 212, 200,   // station 4 X, Y location
    HS + 276, 264,   // station 5 X, Y location
    VS + 340, 328,   // station 6 X, Y location
    HS + 404, 392,   // station 7 X, Y location
    VS + 468, 460    // station 8 X, Y location
};

const unsigned short int level2[] =
{
    256, 768,   // ship starting X, Y location
    8,          // station count
    VS + 480,  60,   // station 1 X, Y location
    HS +420, 120,   // station 2 X, Y location
    VS + 360, 180,   // station 3 X, Y location
    HS +300, 240,   // station 4 X, Y location
    VS + 240, 300,   // station 5 X, Y location
    HS +180, 360,   // station 6 X, Y location
    VS + 120, 420,   // station 7 X, Y location
    HS + 60, 480    // station 8 X, Y location
};

const unsigned short int level3[] =
{
    256, 256,   // ship starting X, Y location
    8,          // station count
    VS + 60, 120,   // station 1 X, Y location
    HS +120,  60,   // station 2 X, Y location
    VS +480, 120,   // station 3 X, Y location
    HS +420,  60,   // station 4 X, Y location
    
    VS + 60, 840,   // station 5 X, Y location
    HS +120, 900,   // station 6 X, Y location
    VS +480, 840,   // station 7 X, Y location
    HS +420, 900    // station 8 X, Y location
};

const unsigned short int level4[] =
{
    256, 768,   // ship starting X, Y location
    8,          // station count
    HS +  60, 512,   // station 1 X, Y location
    HS + 120, 512,   // station 2 X, Y location
    HS + 180, 512,   // station 3 X, Y location
    HS + 240, 512,   // station 4 X, Y location
    HS + 300, 512,   // station 5 X, Y location
    HS + 360, 512,   // station 6 X, Y location
    HS + 420, 512,   // station 7 X, Y location
    HS + 480, 512    // station 8 X, Y location
};

const unsigned short int level5[] =
{
    256, 768,   // ship starting X, Y location
    8,          // station count
    VS +  60, 512,   // station 1 X, Y location
    VS + 120, 512,   // station 2 X, Y location
    VS + 180, 512,   // station 3 X, Y location
    VS + 240, 512,   // station 4 X, Y location
    VS + 300, 512,   // station 5 X, Y location
    VS + 360, 512,   // station 6 X, Y location
    VS + 420, 512,   // station 7 X, Y location
    VS + 480, 512    // station 8 X, Y location
};

const unsigned short int level6[] =
{
    256, 512,   // ship starting X, Y location
    6,          // station count
    VS + 240,   410,    // station 1
    VS + 240,   560,    // station 2
    HS + 176,   450,    // station 3
    HS + 176,   520,    // station 4
    HS + 304,   450,    // station 5
    HS + 304,   520     // station 6
};

const unsigned short int * levels[] =
{
    level1,
    level2,
    level3,
    level4,
    level5,
    level6
};

void InitLevel()
{
    int i;
    int s;
    int l;
    int x;
    int y;
    int a;
    int c;
    int d;
    const unsigned short int *level;
    
//    level = level1;
    level = levels[gLevel];
    
    gInitLevel = 0;
    
    gNonFlickeredSprites = 0;
    for (i=0;i<MAX_SPRITES;i++)
    {
        gSpriteState[i] = 0;
        gSpriteFlickerList1[i] = i;       // init flicker list
    }
    
    gNonFlickeredShots = 0;
    for (i=0;i<MAX_SHOTS;i++)
    {
        gShotState[i] = 0;
        gShotFlickerList1[i] = i;       // init flicker list
    }
    
    // create new random starfield
    gStarfieldXoffset = 0;
    gStarfieldYoffset = 0;
    gStarfieldTopStar = 0;
    for (i=0;i<MAX_STARS;i++)
    {
        STARFIELD_X[i] = (160 * (Random32() & 0xff)) >> 8;          // random 0-159
        STARFIELD_Y[i] = ((Random32() & 0x300) >> 8) + 2;           // random 2-5
        if (MM_TV_TYPE == 2)
            STARFIELD_COLOR[i] = ((14 * (Random32() & 0xff)) >> 8) + 2; // if SECAM use random color 2-15
        else
            STARFIELD_COLOR[i] = ((Random32() & 0xF700) >> 8) + 0x08;   // random color with luma 8-15
    }
    
    // position the player
    l = 0;
    gSpriteState[PLAYER_SPRITE]     = 0x80;
    gSpriteImageID[PLAYER_SPRITE]   = SHIP_ID;
    gSpriteColor[PLAYER_SPRITE]     = gObjectColors[COLOR_SHIP];
    gSpriteX[PLAYER_SPRITE]         = level[l++] << X_FRACTION_BITS; //80;
    gSpriteY[PLAYER_SPRITE]         = level[l++] << Y_FRACTION_BITS; //(DATASTREAM_SIZE - SHIP_HEIGHT) /2;
    gShipDir                        = 0;
    
    gWindowX = gSpriteX[PLAYER_SPRITE] - (80 << X_FRACTION_BITS);
    gWindowY = gSpriteY[PLAYER_SPRITE] - (((DATASTREAM_SIZE - SHIP_HEIGHT) /2) << Y_FRACTION_BITS);
    
    // init station state
    for(i=0;i<8;i++)
        gStationState[i] = 0;

    // position stations
    my_memset(gStationImages, 0, STATION_HEIGHT*8);
    s = level[l++];
    for (i=0;i<s;i++)
    {
        gSpriteState[i]     = 0x80;
        gStationState[i]    = 0x3F; // bits 0-5 control POD states, this turns them all on
        gSpriteImageID[i]   = STATION_HORIZONTAL_ID + (level[l] >> VS_SHIFT);
        gSpriteColor[i]     = gObjectColors[COLOR_STATION];
        gSpriteX[i]         = (level[l++] & STATION_X_MASK) << X_FRACTION_BITS;
        gSpriteY[i]         = level[l++] << Y_FRACTION_BITS;
        my_memcpy((unsigned char *)queue + STATION_IMAGES + i*STATION_HEIGHT,
                  (unsigned char *)flashdata + gImageGraphics[gSpriteImageID[i]],
                  gImageHeight[gSpriteImageID[i]]);
    }
    
    // randomly place mines so they don't overlap
    a = FIRST_STATIONARY_SPRITE;
    d = STATIONARY_DENSITY;  // consider 3, 4 or 5 as a "density/difficulty" option.  Need to make sure enough sprites are available for 5
    for (l=0;l<32;l++)
//    a += d;    for(l=1;l<2;l++)
    {
        for (i=0; i < d; i++)
        {
            do
            {
                x = (( 504 * (Random32() & 0xff)) >> 8) << X_FRACTION_BITS;
                y = ((l * 32) + (Random32() & 15)) << Y_FRACTION_BITS;      // max stationary height is 15, so object should never straddle 2 vertical zones
                c = 0;  // reset collision
                
                
                // player overlap check goes here
                // make sure to add some buffer space around player's position
                if (y + (15 << Y_FRACTION_BITS) >= gSpriteY[PLAYER_SPRITE] - (PLAYER_BUFFER_SPACE << Y_FRACTION_BITS))          // 15 = height of mine
                    if (y <= gSpriteY[PLAYER_SPRITE] + ((PLAYER_BUFFER_SPACE + 17) << Y_FRACTION_BITS))                         // 17 = height of player's ship
                        if (x + (8 << X_FRACTION_BITS) >= gSpriteX[PLAYER_SPRITE] - (PLAYER_BUFFER_SPACE << X_FRACTION_BITS))   // 8 = width of mine
                            if (x <= gSpriteX[PLAYER_SPRITE] + ((9 + PLAYER_BUFFER_SPACE) << X_FRACTION_BITS))                  // 9 = width of ship
                                c = 1;
                
                // stationary overlap check, only checking for X overlap
                if (c == 0)
                    for (s=0;s<i;s++)
                    {
                        if (x + (8 << X_FRACTION_BITS) >= gSpriteX[a+s])
                            if (x <= gSpriteX[a+s] + (8 << X_FRACTION_BITS))
                                c = 1;
                    }

                // station overlap check
                if (c == 0)
                    for (s=0;s<8;s++)
                    {
                        if (gSpriteState[s])
                        {
                            if (y + (15 << Y_FRACTION_BITS) >= gSpriteY[s])             // 15 = height of mine
                                if (y <= gSpriteY[s] + (65 << Y_FRACTION_BITS))         // 65 = height of horizontal station
                                    if (x + (8 << X_FRACTION_BITS) >= gSpriteX[s])      // 8 = width of mine
                                        if (x <= gSpriteX[s] + (32 << X_FRACTION_BITS)) // 32 = width of station
                                            c = 1;
                        }
                    }
            } while (c);
            gSpriteState[a+i]     = 0x80;
            gSpriteX[a+i]         = x;
            gSpriteY[a+i]         = y;
            gSpriteImageID[a+i]   = (Random32() & 3) + ASTEROID_ID;
            gSpriteColor[a+i]     = (gSpriteImageID[a+i] == MINE_ID) ? gObjectColors[COLOR_MINE] : gObjectColors[COLOR_ASTEROID];
        }
        a += d;
    }
//
//    for (i=FIRST_STATIONARY_SPRITE;i<MAX_SPRITES;i++)
//    {
//        gSpriteState[i]     = 0x80;
//        gSpriteX[i]         = (( 504 * (Random32() & 0xff)) >> 8) << X_FRACTION_BITS;
//        gSpriteY[i]         = ((1010 * (Random32() & 0xff00)) >> 8) << Y_FRACTION_BITS;
//        gSpriteImageID[i]   = (Random32() & 3) + ASTEROID_ID;
//        gSpriteColor[i]     = (gSpriteImageID[i] == MINE_ID) ? gObjectColors[COLOR_MINE] : gObjectColor[COLOR_ASTEROID];
//    }
    
    gFormationSize = 4;
    s = ETYPE_ID;
    for (i = FORMATION_SPRITE; i< FORMATION_SPRITE + gFormationSize; i++)
    {
        gSpriteState[i]     = 0x80;
        gSpriteX[i]         = 156 << X_FRACTION_BITS;
        gSpriteY[i]         = (768 + i * 18) << Y_FRACTION_BITS;
        gSpriteImageID[i]   = s;
        gSpriteColor[i]     = (gSpriteImageID[i] == ETYPE_ID) ? gObjectColors[COLOR_E] :
                              (gSpriteImageID[i] == ITYPE_ID) ? gObjectColors[COLOR_I] :
                              (gSpriteImageID[i] == PTYPE_ID) ? gObjectColors[COLOR_P] : gObjectColors[COLOR_SPY];
        s += 2;
    }
    
//    // Pod Alignment Test using 1st station on level
//    s = gSpriteImageID[0] * 7;  // horizontal = 0, vertical = 7
//    for (i=0;i<7;i++)
//    {
//        gSpriteState[13+i]  = 0x80;
//        gSpriteX[13+i]      = gSpriteX[0] + (pod_offset_x[s+i] << X_FRACTION_BITS);
//        gSpriteY[13+i]      = gSpriteY[0] + (pod_offset_y[s+i] << Y_FRACTION_BITS);
//        gSpriteImageID[13+i]= pod_shape[s+i];
//        gSpriteColor[13+i]  = 12;
//    }
    
#ifdef QUAD_TEST

    gShotX[MAX_SHOTS-1] = gShotY[MAX_SHOTS-1] = 80;
    gShotState[MAX_SHOTS-1] = 0x83;

#endif
    
}

void InitDatastreams()
{
    int i;
    // clears datasteams so they can be populated for next frame
    my_memset((unsigned char *)queue + P0_DATASTREAM,               0, DATASTREAM_SIZE);
    my_memset((unsigned char *)queue + HMP0_DATASTREAM,             0, DATASTREAM_SIZE);
    my_memset((unsigned char *)queue + P1_DATASTREAM,               0, DATASTREAM_SIZE);
    my_memset((unsigned char *)queue + HMP1_DATASTREAM,             0, DATASTREAM_SIZE);
    my_memset((unsigned char *)queue + COLOR_HM_DATASTREAM,         0, DATASTREAM_SIZE);
    my_memset((unsigned char *)queue + SIZE_BALL_COLOR_DATASTREAM,  0, DATASTREAM_SIZE);
    my_memset((unsigned char *)queue + MISSILE_BALL_DATASTREAM,     0, DATASTREAM_SIZE);
    my_memset(MISSILE_REPOSITION_ACTIVE,                            0, DATASTREAM_SIZE);
    
    for(i=0;i<DATASTREAM_SIZE;i++)
        gJumpDataStream[i] = NORMAL_KERNEL;
    gJumpDataStream[DATASTREAM_SIZE-1] = KERNEL_DONE;        
}

unsigned int Random32()
{
    // using a 32-bit Galois LFSR as a psuedo random number generator.
    // http://en.wikipedia.org/wiki/Linear_feedback_shift_register#Galois_LFSRs
    
    static unsigned int random = 0x02468ace;
    
    return random = (random >> 1) ^ (unsigned int)(-(random & 1u) & 0xd0000001u);
}

const unsigned char energy_offset_x[] =
{
    10, // Horizontal Station Energy Pod 0
    12, // Horizontal Station Energy Pod 1
    14, // Vertical Station Energy Pod 0
    15  // Vertical Station Energy Pod 1
};

const unsigned char energy_offset_y[] =
{
    31, // Horizontal Station Energy Pod 0
    29, // Horizontal Station Energy Pod 1
    25, // Vertical Station Energy Pod 0
    22  // Vertical Station Energy Pod 1
};

void DisplaySprites()
{
    int i;
    int sprite;
    int shown;
    unsigned char* sprite_shown_ptr;
    unsigned char* sprite_not_shown_ptr;
    int boost;
    int energy_shape;
    int x;
    int y;
    
    sprite_not_shown_ptr = gSpriteFlickerList1;
    sprite_shown_ptr = gSpriteFlickerList2;
    
    PLAYER_0_X = PLAYER_1_X = 0xff;
    FIRST_COLUP0 = FIRST_COLUP1 = 0x0f;
    FIRST_NUSIZ0 = FIRST_NUSIZ1 = 0;    
    
    boost = (MM_TV_TYPE == 2) ? 0 : 2;  // no lumaboost for SECAM
    
    // turn off bit 6 for stations
    // DataStream() will set it on if the station is on-screen
    // so that station shot logic knows which stations can fire.
    for (i=0;i<8;i++)
        gSpriteState[i] &= 0xbf;
    
    for (i=0;i<MAX_SPRITES;i++)
    {
        sprite = gSpriteFlickerList1[i];
        shown = 0;
        
        if (gSpriteState[sprite])   // non-zero value means sprite is in use
        {
            if (gNonFlickeredSprites == sprite_not_shown_ptr)  boost = 0;
            if (DataStream(0, sprite, boost))                  shown = 2; // try to draw using player0
            else if (DataStream(1, sprite, boost))             shown = 1; // try to draw using player1
        }
        if (shown)
        {
            *sprite_shown_ptr = sprite;
            sprite_shown_ptr++;
            if ((sprite < 8) && (gSpriteState[sprite] && SS_STATION_ONSCREEN))
            {
                if (gStationState[sprite] & SS_STATION_OPEN)   // if station is open, display energy field at center
                {
                    
                    gStationState[sprite] ^= SS_ENERGY_FRAME;
                    energy_shape = (gSpriteImageID[sprite] << 1) +
                                   ((gStationState[sprite] & SS_ENERGY_FRAME) >> SS_ENERGY_FRAME_SHIFT);
                    
                    x = (((gSpriteX[sprite] - gWindowX) >> X_FRACTION_BITS) & (WORLD_WIDTH-1)) + energy_offset_x[energy_shape];
                    y = (((gSpriteY[sprite] - gWindowY) >> Y_FRACTION_BITS) & (WORLD_HEIGHT-1)) + energy_offset_y[energy_shape];
                    
                    energy_shape += 4;
                    
                    if (x > (WORLD_WIDTH / 2)) x -= WORLD_WIDTH;
                    if (y > (WORLD_HEIGHT / 2)) y -= WORLD_HEIGHT;
                    if (x > 0 && x < 160 && y > -shot_height[energy_shape] && y < DATASTREAM_SIZE) // 13 = height of tallest energy pod
                    {
                        gShotX[MAX_SHOTS] = x;
                        gShotY[MAX_SHOTS] = y;
                        gShotState[MAX_SHOTS] = energy_shape ;
                        DataStreamShot(shown-1, MAX_SHOTS, 1);
                    }
                }
            }
        }
        else
        {
            *sprite_not_shown_ptr = sprite;
            sprite_not_shown_ptr++;
        }
    }
    gNonFlickeredSprites = sprite_not_shown_ptr;
    my_memcpy(sprite_not_shown_ptr, gSpriteFlickerList2, sprite_shown_ptr-gSpriteFlickerList2);
    
    // if topmost position isn't used then reset to 0 so the 6507 code
    // doesn't take 2 scanlines to reposition the unused player.
    if (PLAYER_0_X == 0xff)        PLAYER_0_X = 0;
    if (PLAYER_1_X == 0xff)        PLAYER_1_X = 0;
}

void DisplayShots()
{
    int i;
    int shot;
    int shown;
    int ds0;
    int ds1;
    unsigned char* shot_shown_ptr;
    unsigned char* shot_not_shown_ptr;
    
    shot_not_shown_ptr = gShotFlickerList1;
    shot_shown_ptr = gShotFlickerList2;
    
    ds0 = FRAME & 1;
    ds1 = 1 - ds0;
    
    MISSILE_0_X = MISSILE_1_X = 0xff;
    
    for (i=0;i<MAX_SHOTS;i++)
    {
        shot = gShotFlickerList1[i];
        shown = 0;
        
        if (gShotState[shot])   // non-zero value means shot is in use
        {
            if (DataStreamShot(ds0, shot, 1))                shown = 1; // try to draw using player0
            else if (DataStreamShot(ds1, shot, 0))           shown = 1; // try to draw using player1
        }
        if (shown)
        {
            *shot_shown_ptr = shot;
            shot_shown_ptr++;
        }
        else
        {
            *shot_not_shown_ptr = shot;
            shot_not_shown_ptr++;
        }
    }
    gNonFlickeredShots = shot_not_shown_ptr;
    my_memcpy(shot_not_shown_ptr, gShotFlickerList2, shot_shown_ptr-gShotFlickerList2);
    
    // if topmost position isn't used then reset to 0 so the 6507 code
    // doesn't take 2 scanlines to reposition the unused missile.
    if (MISSILE_0_X == 0xff)        MISSILE_0_X = 0;
    if (MISSILE_1_X == 0xff)        MISSILE_1_X = 0;
}

//const unsigned char left_mask[] =
//{
//    0xff,
//    0x7f,
//    0x3f,
//    0x1f,
//    0x0f,
//    0x07,
//    0x03,
//    0x01,
//    0x00
//}
//
//const unsigned char right_mask[] =
//{
//    0x00,
//    0x80,
//    0xc0,
//    0xe0,
//    0xf0,
//    0xf8,
//    0xfc,
//    0xfe,
//    0xff
//}
int DataStream(int stream, int index, int lumaboost)
{
    // if possible, copies image to specified datastream
    // returns 1 if successful, 0 if failed.
    static int y;
    static int x;
    static int height;
    static int imageID;
    static int offset;
    static int size;
    static int width;
    
    int hmove;
    int position;
    int i;
    unsigned char * playerstream;
    unsigned char * hmpstream;
    unsigned short int * resptable;
    int mask;
    
    // init only needed when called for stream 0 as same values would have been derived for stream 1
    if (stream == 0)
    {
        imageID = gSpriteImageID[index];
        if (imageID <= QUAD_SPRITES)
        {
            size = 7;
            width = 32;
        }
        else if (imageID <= DOUBLE_SPRITES)
        {
            size = 5;
            width = 16;
        }
        else
        {
            size = 0;
            width = 8;  // no masking needed
        }
        
        x = ((gSpriteX[index] - gWindowX) >> X_FRACTION_BITS) & (WORLD_WIDTH-1);
        y = ((gSpriteY[index] - gWindowY) >> Y_FRACTION_BITS) & (WORLD_HEIGHT-1);
        
        if (x > (WORLD_WIDTH / 2)) x -= WORLD_WIDTH;
        if (y > (WORLD_HEIGHT / 2)) y -= WORLD_HEIGHT;
        
        if (x+width < 8) return 1;    // sprite is offscreen, no need to draw so treat like we did
        if (x > 159) return 1;  // sprite is offscreen, no need to draw so treat like we did

        height = gImageHeight[imageID];
        if (y < 0)
        {
            if (-y >= height)
                return 1;   // sprite is offscreen, no need to draw so treat like we did
            offset = -y;
            height -= offset;
            y = 0;
        }
        else if (y >= DATASTREAM_SIZE)
            return 1;           // sprite is offscreen, no need to draw so treat like we did
        else
            offset = 0;
        
        if (y+height >= DATASTREAM_SIZE)
            height = DATASTREAM_SIZE - y;
        
        // save in state if station is onscreen
        if (index < 8)
            gSpriteState[index] |= SS_STATION_ONSCREEN;
        
        // set stream 0 specific values
        playerstream    = gP0DataStream;
        hmpstream       = gHMP0DataStream;
        resptable       = RESP0_STROBE_TABLE;
    }
    else
    {
        // set stream 1 specific values
        playerstream    = gP1DataStream;
        hmpstream       = gHMP1DataStream;
        resptable       = RESP1_STROBE_TABLE;
    }
    
    if (y < 4)
    {
        // if we're the top-most object, then use PLAYER_X to
        // set the X position if it hasn't already been used
        if (PLAYER_X[1-stream] != 0xff)
            return 0;
        position = 0;
    }
    else
    {
        // all repositions only take 1 scanline.  At this point only player0 and player1
        // repositions could be active so if we see 2 repositions in a row we can abort
        
        if      (gJumpDataStream[y-1] == NORMAL_KERNEL) position = -1;
        else if (gJumpDataStream[y-2] == NORMAL_KERNEL) position = -2;
        // added missile usage for station energy beam
        else if (gJumpDataStream[y-3] == NORMAL_KERNEL) position = -3;
        else if (gJumpDataStream[y-4] == NORMAL_KERNEL) position = -4;
        else                                            return 0;
    }
    
    // check if the player is already in use for the scanlines we need
    for (i = (y + position); i < (y + height) && i < DATASTREAM_SIZE;i++)
    {
        if (hmpstream[i] != 0)  // bit 8 is set for active as player graphics could have gaps in them
            return 0;                                   // abort, player in use
    }
    
    // sprite can go here, so copy it in!
    if (index < 8)
    {
        // copy over station images which are in RAM so they can have pods removed
        my_memcpy(playerstream + y,
                  (unsigned char *)queue + STATION_IMAGES + index*STATION_HEIGHT + offset,
                  height);
    }
    else
    {
        // all other images are in ROM
        my_memcpy(playerstream + y,
                  (unsigned char *)flashdata + gImageGraphics[imageID] + offset,
                  height);
    }
    
    // check for masking on 2x and 4x sprites
    if (width != 8)
    {
        mask = 0xff;
        if (x < 0)                  // mask on left edge
        {
            if (width == 32)
                mask >>= ((6-x) >> 2);
            else
                mask >>= ((6-x) >> 1);
        }
        else if (x + width > 160)   // check if mask needed on right edge
        {
            if (width == 32)
                mask <<= ((x-128) >> 2);
            else
                mask <<= ((x-144) >> 1);
        }
        if (mask != 0xff)
        {
            hmove = 0;
            for (i = (y + position); i < (y + height) && i < DATASTREAM_SIZE;i++)
                hmove += (playerstream[i] &= mask);
            // if mask removes the image, then don't continue processing this sprite.
            // this happens with the Horizontal Station when it gets masked on the
            // left side of the screen.
            if (hmove == 0)
                return 1;
        }
    }

    // "wiggle" large sprites so they look like they're hi-res
    if (imageID <= SHIFTING_SPRITES)
    {
        my_memcpy(hmpstream + y,
                  (unsigned char *)flashdata + gImageShift[imageID] + offset,
                  height);
        x += hmpstream[y] & 0x07;
        hmpstream[y] = 8;   // turn off hmove but set bit 3 for in-use check
    }
    else
    {
        my_memset(hmpstream + y, 8, height);    // set bit 3 in the HMOVE datastream for in-use check
    }
    
    // fat objects get displayed 1 pixel to the right, so compensate
    if (size) x--;
    
    if (position)
    {
        x++;    // reposition routines are off by 1 pixel from 6507 PosObject
        
        if (x <= 4)     // HMOVE with cycle 23 RESPx has issues,
            x += 160;   // this compensates
        
		hmove = 0;
		while (x > 15)
		{
			x -= 15;
			hmove++;
		}
		if (hmove > 10)		hmove = 0;	// prevent a crash
        
        if ((hmove == 0) && (x >= 5))   // HMOVE with cycle 23 RESP1 has issues,
            x -= 5;                     // this compensates
        
		x ^= 7;
		x <<= 4;
        
        gJumpDataStream[y + position]               = resptable[hmove]; // set which cycle to strobe RESPx
        hmpstream[y + position]                     = 8;                // prevent premature sprite reuse
        hmpstream[y + position + 1]                 = x | 8;            // set fine-positioning, prevent premature sprite reuse
        gSizeBallColorDataStream[y + position + 1]  = size;
        gColorHMDataStream[y + position + 1]        = gSpriteColor[index] + lumaboost;
    }
    else
    {
        if (x < 0) x += 160;
        PLAYER_X[1-stream]      = x;
        FIRST_NUSIZ[stream]     = size;
        FIRST_COLUP[stream]     = gSpriteColor[index] + lumaboost;
    }
    
    return 1;
}



int DataStreamShot(int stream, int index, int firstcall)
{
    // if possible, copies shot to specified datastream
    // returns 1 if successful, 0 if failed.
    static int y;
    static int height;
    static int offset;
    static int size;
    
    int hmove;
    int position;
    int x;
    unsigned short int * resptable;
    int missile_on;
    int shape;
    
    // init only needed when for firstcall
    if (firstcall)
    {
        y = gShotY[index];
//        switch (gShotState[index] & 3)
//        {
//            case 0:     height = 1;     size = 0x20;      break;    // horizontal shot
//            case 1:     height = 2;     size = 0x10;      break;    // diagonal shot
//            case 2:     height = 5;     size = 0x00;      break;    // vertical shot
//            case 3:     height = 3;     size = 0x10;      break;    // station shot
//        }
        if (index == MAX_SHOTS) shape = gShotState[MAX_SHOTS];
        else                    shape = gShotState[index] & 3;
        
        height  = shot_height[shape];
        size    = shot_size[shape];

        if (y < 0)
        {
            if (-y >= height)
                return 1;   // sprite is offscreen, no need to draw so treat like we did
            offset = -y;
            height -= offset;
            y = 0;
        }
        else
            offset = 0;
        
        if (y+height >= DATASTREAM_SIZE)
            height = DATASTREAM_SIZE - y;
    }
    if (stream == 0)
    {
        // set stream 0 specific values
        missile_on      = 0x04;
        resptable       = RESM0_STROBE_TABLE;
    }
    else
    {
        // set stream 1 specific values
        missile_on      = 0x08;
        resptable       = RESM1_STROBE_TABLE;
    }
    
    if (y < 4)
    {
        // if we're the top-most object, then use MISSILE_X to
        // set the X position if it hasn't already been used
        if (MISSILE_X[1-stream] != 0xff)
            return 0;
        position = 0;
    }
    else
    {
        // all repositions only take 1 scanline.
        if      (gJumpDataStream[y-1] == NORMAL_KERNEL) position = -1;
        else if (gJumpDataStream[y-2] == NORMAL_KERNEL) position = -2;
        else if (gJumpDataStream[y-3] == NORMAL_KERNEL) position = -3;
        else if (gJumpDataStream[y-4] == NORMAL_KERNEL) position = -4;
        else                                            return 0;
    }
    
    // check if the missile is already in use for the scanlines we need
    for (x = (y + position); x < (y + height) && x < DATASTREAM_SIZE;x++)
    {
        if (MISSILE_REPOSITION_ACTIVE[x] & missile_on)
            return 0;
    }
    
    // missile can go here, so copy it in!
    for (x=y; x<(y+height) ;x++)
        gMissileBallDataStream[x]     |= missile_on;   // turn on the missile
    
    // flag that a reposition is active
    for (x=y+position;x<(y+height) ;x++)
        MISSILE_REPOSITION_ACTIVE[x] |= missile_on;        
    
    x = gShotX[index] + 1; // missiles' X is off by 1, so compensate
    
    if (position)
    {
        x++;    // reposition routines are off by 1 pixel from 6507 PosObject
        
        if (x <= 4)     // HMOVE with cycle 23 RESPx has issues,
            x += 160;   // this compensates
        
		hmove = 0;
		while (x > 15)
		{
			x -= 15;
			hmove++;
		}
		if (hmove > 10)		hmove = 0;	// prevent a crash
        
        if ((hmove == 0) && (x >= 5))   // HMOVE with cycle 23 RESP1 has issues,
            x -= 5;                     // this compensates
        
		x ^= 7;
		x <<= 4;
        
        gJumpDataStream[y + position]               = resptable[hmove]; // set which cycle to strobe RESPx
        gColorHMDataStream[y + position +1]         = x;
        gSizeBallColorDataStream[y + position + 1]  = size;
    }
    else
    {
        MISSILE_X[1-stream]     = x;
        FIRST_NUSIZ[stream]    |= size;
    }
    
    return 1;
}


void DisplayStars()
{
    int i;
    int x;
    int y;
    int hmove;
    int blink;
    
    y = gStarfieldYoffset >> STARFIELD_FRAC_BITS;
    blink = ((FRAME >> 5) + gStarfieldTopStar) & 7;
    i = gStarfieldTopStar;
    
    while (y < DATASTREAM_SIZE - 1)
    {
        if (blink && (gJumpDataStream[y]  == NORMAL_KERNEL))
        {
            x = STARFIELD_X[i] + (gStarfieldXoffset >> STARFIELD_FRAC_BITS);
            if (x > 159) x-= 160;
            if (x <= 4)     // HMOVE with cycle 23 RESPx has issues,
                x += 160;   // this compensates
            
            hmove = 0;
            while (x > 15)
            {
                x -= 15;
                hmove++;
            }
            if (hmove > 10)		hmove = 0;	// prevent a crash
            
            if ((hmove == 0) && (x >= 5))   // HMOVE with cycle 23 RESP1 has issues,
                x -= 5;                     // this compensates
            
            x ^= 7;
            x <<= 4;
            
            gJumpDataStream[y]          = RESBL_STROBE_TABLE[hmove]; // set which cycle to strobe RESPx
            gColorHMDataStream[y + 1]       = x;
            gSizeBallColorDataStream[y + 1] = STARFIELD_COLOR[i];
            gMissileBallDataStream[y+1]     |= 2;   // turn on the star
        }
        y += STARFIELD_Y[i];
        blink++;
        blink &= 7;
        
        i++;
        i &= 127;
    }
}

const unsigned char joystick_direction[] = { 9, 9, 9, 9, 9, 3, 1, 2, 9, 5, 7, 6, 9, 4, 0, 9};
//  7 0 1
//  6   2
//  5 4 3

const signed char joy_dir_x[] = { 0, 1, 1, 1, 0,-1,-1,-1 };
const signed char joy_dir_y[] = {-1,-1, 0, 1, 1, 1, 0,-1 };
const unsigned char ship_x_adj[] = { 0, 1, 1, 1, 0, 0, 0, 0 };
//const unsigned char ship_y_adj[] = { 0, 1, 0, 0, 0, 0, 0, 0 };
const unsigned char shot_shape[] = { 2, 1, 0, 1};

const signed char shot_offset_x[] = { 4, 6, 4,  6,  4,  0, 0,-1};
const signed char shot_offset_y[] = {-3, 4, 7, 11, 14, 11, 7, 3};

void ProcessJoystick()
{
    int i;
    int x;
    int y;
    int free_shot[2];
    
    int direction;
    
#ifdef QUAD_TEST
    direction = joystick_direction[SWCHA & 0x0f];
    
    if (direction < 9)
    {
        gShotX[MAX_SHOTS-1] += joy_dir_x[direction];
        gShotY[MAX_SHOTS-1] += joy_dir_y[direction];
        
        if (gShotX[MAX_SHOTS-1] < 8)    gShotX[MAX_SHOTS-1] = 159;
        if (gShotX[MAX_SHOTS-1] > 159)  gShotX[MAX_SHOTS-1] = 8;
        if (gShotY[MAX_SHOTS-1] == 255)  gShotY[MAX_SHOTS-1] = DATASTREAM_SIZE-1;
        if (gShotY[MAX_SHOTS-1] >= DATASTREAM_SIZE)  gShotY[MAX_SHOTS-1] = 0;
    }
#endif
    
    if (SWCHB & 0x80)   // Right Difficulty A = stationary ship
            return;
    
    x = y = 0;
    
    direction = joystick_direction[SWCHA >> 4];
    
    if ((gSpriteState[PLAYER_SPRITE] & 0xC0) == 0x80)
    {
        if (direction < 8)
            gShipDir = direction;
        gSpriteImageID[PLAYER_SPRITE]   = SHIP_ID + gShipDir;
    }
    else
    {
        // player's exploding - disable firebutton
        INPT4 = 0xff;
    }
    
    x = joy_dir_x[gShipDir];
    y = joy_dir_y[gShipDir];
    gWindowX += x << X_FRACTION_BITS;
    gWindowY += y << Y_FRACTION_BITS;
    gSpriteX[PLAYER_SPRITE] = gWindowX + ((80 + ship_x_adj[gShipDir]) << X_FRACTION_BITS);
    gSpriteY[PLAYER_SPRITE] = gWindowY + ((((DATASTREAM_SIZE - SHIP_HEIGHT) /2) /*+ ship_y_adj[gShipDir]*/ ) << Y_FRACTION_BITS);

    // player shots don't shift with the window, so compenstate their X Y values
    for (i=4;i<MAX_SHOTS;i++)
    {
        if (gShotState[i])
        {
            gShotX[i] -= x;
            gShotY[i] -= y;
        }
    }
    
    // update starfield based on player's movement
    gStarfieldXoffset -= x;
    gStarfieldYoffset -= y;
    
    if (gStarfieldXoffset < 0 ) gStarfieldXoffset = 159 << STARFIELD_FRAC_BITS;
    if (gStarfieldXoffset >> STARFIELD_FRAC_BITS == 160 ) gStarfieldXoffset = 0;
    
    if (y > 0)
    {
        if (gStarfieldYoffset < 0)
        {
            gStarfieldYoffset = STARFIELD_Y[gStarfieldTopStar] << STARFIELD_FRAC_BITS;
            gStarfieldTopStar++;
            gStarfieldTopStar &= 127;
        }
    }
    else
    {
        if (gStarfieldYoffset >> STARFIELD_FRAC_BITS == STARFIELD_Y[(gStarfieldTopStar - 1) & 127])
        {
            gStarfieldYoffset = 0;
            gStarfieldTopStar--;
            gStarfieldTopStar &= 127;
        }
    }
    

    // gun processing
    if (gGunCoolDown)   gGunCoolDown--;
    if (!(INPT4 & 0x80))  // fire down
    {
        x = 0;
        for (i=0;i<4 && x < 2;i++)
        {
            if (gShotState[i] == 0)
            {
                free_shot[x] = i;
                x++;
            }
        }
        
        if (x == 2 && gGunCoolDown == 0)
        {
            TriggerSoundEffect(SOUND_PLAYER_SHOOTS_ID);
            gGunCoolDown = GUN_COOL_DOWN;
            for (i=0;i<2;i++)
            {
                x = free_shot[i];
                gShotState[x] = 0x80 + (((gShipDir + i*4) & 7) << SHOT_DIRECTION_SHIFT)+ shot_shape[gShipDir & 3];
                gShotX[x] = ((gSpriteX[PLAYER_SPRITE] - gWindowX) >> X_FRACTION_BITS) + shot_offset_x[(gShipDir + i * 4) & 7];
                gShotY[x] = ((gSpriteY[PLAYER_SPRITE] - gWindowY) >> Y_FRACTION_BITS) + shot_offset_y[(gShipDir + i * 4) & 7];
            }
        }
    }
}

void FixSizes()
{
    int i;
    unsigned char player[2];
    unsigned char missile[2];
    
    // get starting sizes
    player[0]   = FIRST_NUSIZ0 & 7;
    player[1]   = FIRST_NUSIZ1 & 7;
    missile[0]  = FIRST_NUSIZ0 & 0x30;
    missile[1]  = FIRST_NUSIZ1 & 0x30;
    
    for(i=0;i<DATASTREAM_SIZE-1;i++)
    {
        // look for Jumps for player or missile repositioning
        if (gJumpDataStream[i] >= FIRST_P0_JUMP && gJumpDataStream[i] <= LAST_M1_JUMP)
        {
            if (gJumpDataStream[i] <= LAST_P1_JUMP) // test for player jumps
            {
                if (gJumpDataStream[i] <= LAST_P0_JUMP)
                {
                    player[0] = gSizeBallColorDataStream[i+1];    // get new player size
                    gSizeBallColorDataStream[i+1] |= missile[0];  // merge in current missile size
                }
                else
                {
                    player[1] = gSizeBallColorDataStream[i+1];    // get new player size
                    gSizeBallColorDataStream[i+1] |= missile[1];  // merge in current missile size
                }
            }
            else                                    // it's a missile jump
            {
                if (gJumpDataStream[i] <= LAST_M0_JUMP)
                {
                    missile[0] = gSizeBallColorDataStream[i+1];   // get new missile size
                    gSizeBallColorDataStream[i+1] |= player[0];   // merge in current player size
                }
                else
                {
                    missile[1] = gSizeBallColorDataStream[i+1];   // get new missile size
                    gSizeBallColorDataStream[i+1] |= player[1];   // merge in current player size
                }
            }
        }
    }
}

const unsigned char digit_stream[] =
{
    PF1R, // + (RADAR_HEIGHT - PF_DIGIT_HEIGHT + 1) / 2,   // ones digit
    PF0R, // + (RADAR_HEIGHT - PF_DIGIT_HEIGHT + 1) / 2,   // tens digit
    PF2L, // + (RADAR_HEIGHT - PF_DIGIT_HEIGHT + 1) / 2,   // hundreds digit
    PF2L, // + (RADAR_HEIGHT - PF_DIGIT_HEIGHT + 1) / 2,   // thousands digit
    PF1L, // + (RADAR_HEIGHT - PF_DIGIT_HEIGHT + 1) / 2,   // ten thousands digit
    PF1L, // + (RADAR_HEIGHT - PF_DIGIT_HEIGHT + 1) / 2,   // hundred thousands digit
    PF2R, // + (RADAR_HEIGHT - PF_DIGIT_HEIGHT + 1) / 2    // reserve lives digit
};

const unsigned char digit_offset[] =
{
    5,      // ones digit
    4,      // tens digit
    4,      // hundreds digit
    0,      // thousands digit
    1,      // ten thousands digit
    5,      // hundred thousands digit
    3       // reserve lives digit
};

const unsigned char digit_flipped[] =
{
    0,      // ones digit
    1,      // tens digit
    1,      // hundreds digit
    1,      // thousands digit
    0,      // ten thousands digit
    0,      // hundred thousands digit
    1       // reserve lives digit
};

const unsigned char flip7[] =
{
    0,  // 0 000 -> 000
    4,  // 1 001 -> 100
    2,  // 2 010 -> 010
    6,  // 3 011 -> 110
    1,  // 4 100 -> 001
    5,  // 5 101 -> 101
    3,  // 6 110 -> 011
    7   // 7 111 -> 111
};


void ShowDigit(int digit, int position)
{
    unsigned char temp;
    int i;
    
    for (i=0;i<PF_DIGIT_HEIGHT;i++)
    {
        temp = PF_DIGIT0[(digit * PF_DIGIT_HEIGHT) + i];
        if (digit_flipped[position]) temp = flip7[temp];
        temp <<= digit_offset[position];
        gPlayfield[digit_stream[position] + i] |= temp;
    }
}



void PrepScoreRadar()
{
    int i;
    int x;
    int y;
    int score;
    for (i=0;i<RADAR_HEIGHT;i++)
    {
        gPlayfield[i + PF0L] = 0x00;
        gPlayfield[i + PF1L] = 0x00;
        gPlayfield[i + PF2L] = 0x00;
        gPlayfield[i + PF0R] = 0x00;
        gPlayfield[i + PF1R] = 0x00;
        gPlayfield[i + PF2R] = 0x00;
        gP0DataStream[DATASTREAM_SIZE + 1 + i] = 0;
        gP1DataStream[DATASTREAM_SIZE + 1 + i] = 0;
    }
    
    if (SWCHB & 0x40)
    {
        // left difficulty = A, show diagnostics
        ShowDigit(PF_DIGIT_V, 5);
        ShowDigit(VB_TIME >> 4, 4);
        ShowDigit(VB_TIME & 15, 3);
        ShowDigit(PF_DIGIT_O, 2);
        ShowDigit(OS_TIME >> 4, 1);
        ShowDigit(OS_TIME & 15, 0);
    }
    else
    {
        score = gScore;
        for (i=0;i<6;i++)
        {
            ShowDigit(score & 0x0f, i);
            score >>= 4;
            if (score == 0) break;
        }
    }
    
    // prep radar frame
    gP0DataStream[DATASTREAM_SIZE +  1] = 0xc0;  // top left
    gP0DataStream[DATASTREAM_SIZE +  2] = 0x80;  // top left
    gP1DataStream[DATASTREAM_SIZE +  1] = 0x03;  // top right
    gP1DataStream[DATASTREAM_SIZE +  2] = 0x01;  // top right
    gP0DataStream[DATASTREAM_SIZE + 19] = 0x80;  // bottom left
    gP0DataStream[DATASTREAM_SIZE + 20] = 0xc0;  // bottom left
    gP1DataStream[DATASTREAM_SIZE + 19] = 0x01;  // bottom right
    gP1DataStream[DATASTREAM_SIZE + 20] = 0x03;  // bottom right
    
    // show reserve lives, or X if over 9
    i = (gReserveLives <= 9) ? gReserveLives : PF_DIGIT_X;
    ShowDigit(i, 6);             // position 6 = reserve lives
    
    
    // test condition color
    if (FRAME == 0)
    {
        if (gCondition == 2)
            gCondition = 0;
        else
            gCondition++;
    }
    
    // set condition color
    i = gObjectColors[COLOR_CONDITION_GREEN + gCondition];
    if ((gCondition == 2) && (FRAME & 0x10))  i = 0;
    RADAR_CONTROL[0] = i;

    // update station location on radar display
    for (i=0;i<8;i++)
    {
        if (gSpriteState[i])
        {
            x = gSpriteX[i] >> (X_FRACTION_BITS + RADAR_SHIFT);
            y = gSpriteY[i] >> (Y_FRACTION_BITS + RADAR_SHIFT);
            gP1DataStream[DATASTREAM_SIZE + 3 +  y] |= 1 << (7-x);
        }
    }

    // update formation location on radar display
    if (gSpriteState[FORMATION_SPRITE])
    {
        x = gSpriteX[FORMATION_SPRITE] >> (X_FRACTION_BITS + RADAR_SHIFT);
        y = gSpriteY[FORMATION_SPRITE] >> (Y_FRACTION_BITS + RADAR_SHIFT);
        gP0DataStream[DATASTREAM_SIZE + 3 + y] |= 1 << (7-x);
    }
    
    // update player location on radar display
    if ((FRAME & 8) == 8)
    {
        x = -(gSpriteX[PLAYER_SPRITE] >> (X_FRACTION_BITS + RADAR_SHIFT));
//        x += 2;                 // offset for ball                             - NOT NEEDED
        x <<= 4;                // change to HMOVE value
        RADAR_CONTROL[1] = x;   // adjust ball position
        
        y = gSpriteY[PLAYER_SPRITE] >> (Y_FRACTION_BITS + RADAR_SHIFT);
        gPlayfield[y + PF0L] |= 2;
    }
}

//const signed char dir_x[] = { 1, 1, 0,-1,-1,-1, 0, 1 };
//const signed char dir_y[] = { 0, 1, 1, 1, 0,-1,-1,-1 };

void MoveShots()
{
    int i;
    int speed;
    speed = 3;
#ifdef QUAD_TEST
    for (i=0;i<MAX_SHOTS - QUAD_TEST;i++)
#else
    for (i=0;i<MAX_SHOTS;i++)
#endif
    {
        if (i == 4) speed = 1;
            
        if (gShotState[i])
        {
            gShotX[i] += speed*joy_dir_x[(gShotState[i] & SHOT_DIRECTION_MASK) >> SHOT_DIRECTION_SHIFT];
            gShotY[i] += speed*joy_dir_y[(gShotState[i] & SHOT_DIRECTION_MASK) >> SHOT_DIRECTION_SHIFT];
            
            if (gShotX[i] < 8 || gShotX[i] >= 160 || gShotY[i] < 0 || gShotY[i] >= DATASTREAM_SIZE)
                gShotState[i] = 0;
        }
    }
}

void TriggerSoundEffect(int SoundID)
{
    int channel;
    
    if (flashdata[SOUND_VOL + ACTIVE_SOUND_0] == 0)         channel = 0;
    else if (flashdata[SOUND_VOL + ACTIVE_SOUND_1] == 0)    channel = 1;
    else if (SoundID > ACTIVE_SOUND_0)                      channel = 0;
    else if (SoundID > ACTIVE_SOUND_1)                      channel = 1;
    else                                                    return;
    
    ACTIVE_SOUND[channel] = SoundID;
    SOUND_DURATION[channel] = flashdata[SOUND_FREQ + SoundID] >> 5;
}

void UpdateSound()
{
    int channel;
    for (channel = 0;channel < 2; channel++)
    {
        AUDV[channel] = flashdata[SOUND_VOL + ACTIVE_SOUND[channel]];
        AUDC[channel] = AUDV[channel] >> 4;
        AUDF[channel] = flashdata[SOUND_FREQ + ACTIVE_SOUND[channel]];
        
        if (AUDV[channel])
        {
            if (SOUND_DURATION[channel])
            {
                SOUND_DURATION[channel]--;
            }
            else
            {
                ACTIVE_SOUND[channel]--;
                SOUND_DURATION[channel] = flashdata[SOUND_FREQ + ACTIVE_SOUND[channel]] >> 5;
            }
        }
//        if (DEMO_MODE)
//            AUDV[channel] = (AUDV[channel] & 0x0f) >> 2;
    }
}


void ProcessMenuJoystick()
{
    if (!(INPT4 & 0x80))	// firebutton down
	{
        MM_FIRE_DOWN = 1;
	}
	else
	{
		if (MM_FIRE_DOWN)	// if fire button was released
		{
			MM_FIRE_DOWN = 0;
			if (MM_SELECTED_OPTION == MENU_START_ID)
                NewGame();  // start game when firebutton released
		}
	}

    // joystick timer controls auto-changing of values when fire is held
	if (((SWCHA & 0xF0) == 0xF0) && (!MM_FIRE_DOWN)) // if nothing is down
	{
		MM_JOYSTICK_TIMER = 0;
	}
	else
	{
		MM_JOYSTICK_TIMER++;
		if (MM_JOYSTICK_TIMER == 16)
			MM_JOYSTICK_TIMER = 0;
	}
	if (MM_JOYSTICK_TIMER == 1)
	{
		if (!(SWCHA & 0x10) && (!MM_FIRE_DOWN))			// if joystick up and no fire button then change selected option
		{
            if (MM_SELECTED_OPTION ==  MENU_FIRST_OPTION_ID)
                MM_SELECTED_OPTION = MENU_START_ID;
            else
                MM_SELECTED_OPTION--;
            
            return;
		}
        if (!(SWCHA & 0x20) && (!MM_FIRE_DOWN))	// if joystick down and no firebutton then change selected option
		{
            if (MM_SELECTED_OPTION == MENU_START_ID)
                MM_SELECTED_OPTION = MENU_FIRST_OPTION_ID;
            else
                MM_SELECTED_OPTION++;
            
            return;
		}
        if (!(SWCHA & 0x40))	// joystick left
		{
            if (MM_SELECTED_OPTION != MENU_START_ID)
                TriggerSoundEffect(SOUND_MENU_BLIP_ID);
			switch (MM_SELECTED_OPTION)
			{
                case MENU_LEVEL_ID:
                    if (MM_INITIAL_LEVEL == 0)
                        MM_INITIAL_LEVEL = (sizeof(levels) / sizeof(levels[0]))-1;
                    else
                        MM_INITIAL_LEVEL--;
                    break;
                    
                case MENU_TV_TYPE_ID: // TV Type: [0]=NTSC, 1=PAL, 2=SECAM
                    if (MM_TV_TYPE == 0)
                        MM_TV_TYPE = 2;
                    else
                        MM_TV_TYPE--;
                    break;
            }
        }
        else if (!(SWCHA & 0x80) || (MM_FIRE_DOWN))	// joystick right
		{
            if (MM_SELECTED_OPTION != MENU_START_ID)
                TriggerSoundEffect(SOUND_MENU_BLIP_ID);
			switch (MM_SELECTED_OPTION)
			{
                case MENU_LEVEL_ID:
                    if (MM_INITIAL_LEVEL == (sizeof(levels) / sizeof(levels[0]))-1)
                        MM_INITIAL_LEVEL = 0;
                    else
                        MM_INITIAL_LEVEL++;
                    break;
                    
                case MENU_TV_TYPE_ID: // TV Type: [0]=NTSC, 1=PAL, 2=SECAM
                    if (MM_TV_TYPE == 2)
                        MM_TV_TYPE = 0;
                    else
                        MM_TV_TYPE++;
                    break;
            }
        }
    }
}

void MoveFormation()
{
    int i;
    int x;
    int y;
    int s;
    if (FRAME & 0x80)
    {
        s = ETYPE_ID;       // move up
        x = 0;
        y = -1 << Y_FRACTION_BITS;
    }
    else
    {
        s = ETYPE_ID + 1;   // move right
        x = 1 << X_FRACTION_BITS;
        y = 0;
    }
    
    for (i = FORMATION_SPRITE; i< FORMATION_SPRITE + gFormationSize; i++)
    {
        gSpriteX[i]         += x;
        gSpriteY[i]         += y;
        gSpriteImageID[i]   = s;
        s += 2;
    }
}

void NextLevel()
{
    if (gLevel == (sizeof(levels) / sizeof(levels[0]))-1)
        gLevel = 0;
    else
        gLevel++;
    gInitLevel=1;
}

const unsigned char pod_off_y[] =
{
    0,  // Horizontal Pod 0
    9,  // Horizontal Pod 1
    40, // Horizontal Pod 2
    53, // Horizontal Pod 3
    41, // Horizontal Pod 4
    9,  // Horizontal Pod 5
    0,  // Vertical Pod 0
    0,  // Vertical Pod 1
    17, // Vertical Pod 2
    37, // Vertical Pod 3
    37, // Vertical Pod 4
    17  // Vertical Pod 5
};

#define OFF_MASK_SIZE 23
const unsigned char pod_off_mask[] =
{
    // AND - bits that are 0 will be turned off.
    // Pad rows with fewer than OFF_MASK_SIZE entries with 0xFF
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x10, 0x20, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // Horizontal Pod 0
    0x60, 0xff, 0x60, 0xff, 0xff, 0xff, 0xfc, 0xf8, 0xfc, 0xf8, 0xfc, 0xfc, 0xf8, 0xfc, 0xfc, 0xf8, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // Horizontal Pod 1
    0xf8, 0xfc, 0xf8, 0xf8, 0xfc, 0xf8, 0xfc, 0xfc, 0xf8, 0xfc, 0xf8, 0xff, 0xff, 0xf0, 0xff, 0xf0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // Horizontal Pod 2
    0x20, 0x10, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // Horizontal Pod 3
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0xff, 0xff, 0xff, 0x3f, 0x3f, 0x3f, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // Horizontal Pod 4
    0x3f, 0xff, 0x3f, 0xff, 0xff, 0xff, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // Horizontal Pod 5
    0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x2f, 0x0f, 0x4f, 0xff, 0xff, 0xff, 0xff, 0x0f, 0xff, 0x0f, 0xFF, 0xFF, 0xFF, // Vertical Pod 0
    0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf8, 0xff, 0xf4, 0xff, 0xff, 0xff, 0xff, 0xf0, 0xff, 0xf0, 0xFF, 0xFF, 0xFF, // Vertical Pod 1
    0xf0, 0xff, 0xf0, 0xff, 0xf0, 0xff, 0xfc, 0xfe, 0xfe, 0xfe, 0xfc, 0xfe, 0xfc, 0xfc, 0xfe, 0xfc, 0xfe, 0xff, 0xff, 0xff, 0xf0, 0xff, 0xf0, // Vertical Pod 2
    0xf0, 0xff, 0xf0, 0xff, 0xff, 0xff, 0xff, 0xf4, 0xf8, 0xf8, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xFF, 0xFF, 0xFF, // Vertical Pod 3
    0x0f, 0xff, 0x0f, 0xff, 0xff, 0xff, 0xff, 0x4f, 0x3f, 0x3f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0x0f, 0xFF, 0xFF, 0xFF, // Vertical Pod 4
    
    0x0f, 0xff, 0x0f, 0xff, 0x7f, 0xff, // 22
    0x7f, 0x7f, 0x7f, 0x7f, 0x7f, 0x7f, // 28
    0x1f, 0x7f, 0x7f, 0x7f, 0x7f, 0xff, // 34
    0xff, 0xff, 0x0f, 0xff, 0x0f
};

const unsigned char pod_on_y[] =
{
    0,  // Horizontal Pod 0
    0,  // Horizontal Pod 1
    0,  // Horizontal Pod 2
    0,  // Horizontal Pod 3
    0,  // Horizontal Pod 4
    0,  // Horizontal Pod 5
    11, // Vertical Pod 0
    0,  // Vertical Pod 1
    0,  // Vertical Pod 2
    0,  // Vertical Pod 3
    0,  // Vertical Pod 4
    0   // Vertical Pod 5
};

#define ON_MASK_SIZE 1
const unsigned char pod_on_mask[] =
{
    // OR - bits that are 1 will be turned on.
    // Pad rows with fewer than ON_MASK_SIZE entries with 0x00
    0x00,   // Horizontal Pod 0
    0x00,   // Horizontal Pod 1
    0x00,   // Horizontal Pod 2
    0x00,   // Horizontal Pod 3
    0x00,   // Horizontal Pod 4
    0x00,   // Horizontal Pod 5
    0x20,   // Vertical Pod 0
    0x00,   // Vertical Pod 1
    0x00,   // Vertical Pod 2
    0x00,   // Vertical Pod 3
    0x00,   // Vertical Pod 4
    0x00    // Vertical Pod 5
};


void PodHit(int station, int pod)
{
    int i;
    int y;
    
    // pods are numbered 0-5 in this arrangement around the station cores:
    //  0     0 1
    // 5 1   5   2  <- Vertical
    // 4 2    4 3
    //  3  <---------- Horizontal
    
    if (pod < 0 || pod >= 6)
        return;
    
    pod += gSpriteImageID[station] * 6;     // ID = 0 for Horizontal, 1 for Vertical
    
    station *= STATION_HEIGHT;
    
    // turn off pixels
    y =pod_off_y[pod];
    for (i=0;i<OFF_MASK_SIZE && y+i < STATION_HEIGHT;i++)
        gStationImages[station + i + y]    &= pod_off_mask[i + pod * OFF_MASK_SIZE];
    
    // turn on pixels
    y =pod_on_y[pod];
    for (i=0;i<ON_MASK_SIZE && y+i < STATION_HEIGHT;i++)
        gStationImages[station + i + y]     |= pod_on_mask[i + pod * ON_MASK_SIZE];
    
}

const unsigned char energy_offset[] = {  28, 21};
const unsigned char energy_size[]   = {   9, 15};
const unsigned char energy_mask[]   = {0xCF, 0xE7};

void OpenStation(int station, int amount)
{
    int i;
    int offset;
    int size;
    int mask;
    
    offset  = energy_offset[gSpriteImageID[station]] + station * STATION_HEIGHT;
    size    = energy_size[gSpriteImageID[station]];
    mask    = energy_mask[gSpriteImageID[station]];
    if (amount < size)
        gStationState[station] &= (0xFF - SS_STATION_OPEN);  // turn off open flag
    else
        gStationState[station] |= SS_STATION_OPEN;  // turn on open flag
    
    for(i=0;i<size && i < amount;i++)
        gStationImages[offset + i] &= mask;
}

void PrepStationGraphics()
{
    static int open = 1;
    static int amount = 0;
    
    int i;
    int j;
    int count;
    
    // test of open/close of station
    if ((FRAME & 31) == 0)
    {
        if (open)
        {
            if (amount < 16)
                amount++;
            else
                open = 0;
        }
        else
        {
            if (amount)
                amount--;
            else
                open = 1;
        }
    }
    
    count = 0;
    for (i=0;i<8;i++)
    {
        if (gSpriteState[i])
        {
            count++;
            // fresh copy of station image
            my_memcpy((unsigned char *)queue + STATION_IMAGES + i*STATION_HEIGHT,
                      (unsigned char *)flashdata + gImageGraphics[gSpriteImageID[i]],
                      gImageHeight[gSpriteImageID[i]]);
            
            // manipulate graphics for energy pod
            OpenStation(i, amount);
            
            for (j=0;j<6;j++)
            {
                if ((gStationState[i] & (1 << j)) == 0)
                    PodHit(i, j);
            }
        }
    }
    if (count == 0)
        NextLevel();
}

void StationShots()
{
    int i;
    int f;
    int s;
    
    s = -1;
    for (i=0;i<8;i++)
    {
        if (gSpriteState[i] & STATION_COOL_DOWN)               // cooldown active
            gSpriteState[i]--;
        else if ((gSpriteState[i] & 0x5f) == SS_STATION_ONSCREEN)    // if onscreen & cooldown then we can shoot
            s = i;
    }
    
    f = 0;
#ifdef QUAD_TEST
    for (i=4;i<MAX_SHOTS - QUAD_TEST;i++)
#else
    for (i=4;i<MAX_SHOTS;i++)
#endif
        if (gShotState[i] == 0)
            f = i;
    
    if (f && s != -1)
    {
        gSpriteState[s] |= STATION_COOL_DOWN;
        gShotState[f] = 0x80 + ((Random32() & 7) << SHOT_DIRECTION_SHIFT) + 3;
        gShotX[f] = ((gSpriteX[s] - gWindowX) >> X_FRACTION_BITS) + 16;
        gShotY[f] = ((gSpriteY[s] - gWindowY) >> Y_FRACTION_BITS) + 32;
    }
}

unsigned char NTSCtoPAL[16] =
{
    // SeaGtGruff's conversion values from this post at AtariAge:
    // http://www.atariage.com/forums/topic/165424-modify-colour-palette/page__p__2043124#entry2043124
    0x00, 0x20, 0x20, 0x40, 0x60, 0x80, 0xa0, 0xc0,
    0xd0, 0xb0, 0x90, 0x70, 0x50, 0x30, 0x30, 0x20
};

int ColorPAL(int color)
{
    return NTSCtoPAL[color>>4] +    // chroma value
    (color & 0x0f);          // luma value
}


unsigned char NTSCtoSECAM[16] =
{
    0x0e,   // 0 = white            e = white
    0x0c,   // 1 = yellow           c = yellow
    0x0c,   // 2 = yellowish orange c = yellow
    0x04,   // 3 = reddish orange   4 = red
    0x04,   // 4 = red              4 = red
    0x06,   // 5 = reddish purple   6 = purple
    0x06,   // 6 = purple           6 = purple
    0x06,   // 7 = bluish purple    6 = purple
    0x02,   // 8 = blue             2 = blue
    0x02,   // 9 = blue             2 = blue
    0x0a,   // a = bluish cyan      a = cyan
    0x0a,   // b = cyan             a = cyan
    0x08,   // c = green            8 = green
    0x08,   // d = green            8 = green
    0x08,   // e = olive green      8 = green
    0x0c    // f = brown            c = yellow
};

int ColorSECAM(int color)
{
    if (color < 2)
        return 0;   // return black for 0 or 1
    else
        return NTSCtoSECAM[color>>4];
}

int ColorConvert(int color)
{
    if (MM_TV_TYPE == PAL)
        return ColorPAL(color);
    else if (MM_TV_TYPE == SECAM)
        return ColorSECAM(color);
    else
        return color;
}

const unsigned char double_nybble[] = { 0x00, 0x03, 0x0C, 0x0F,
    0x30, 0x33, 0x3C, 0x3F,
    0xC0, 0xC3, 0xCC, 0xCF,
    0xF0, 0xF3, 0xFC, 0xFF};

int DoubleBits(int bits)
{
    return (double_nybble[(bits & 0xf0) >> 4] << 8) + double_nybble[bits & 0x0f];
}

int Collision(int image1_id, int left1, int top1, int sprite_index)
{
    // test if image1_id at position left1, top1 has collided with sprite_index
    // image1_id    = index for graphics & height
    // left1        = top-left corner of sprite, already had >> X_FRACTION_BITS done
    // top1         = top-left corner of sprite, already had >> Y_FRACTION_BITS done
    // sprite_index = index to check against
    //
    //  NOTE: calling routine must make sure sprite_index is for an active sprite
    
    int right1;
    int bottom1;
    
    int left2;
    int top2;
    int right2;
    int bottom2;
    int image2_id;
    int double2;
    
    int top3;
    int bottom3;
    
    int j;
    
    unsigned char * image1;
    unsigned char * image2;
    
    int y_overlap;
    int x_shift;
    
    right1 = left1 + 8;
    bottom1 = top1 + gImageHeight[image1_id];
    
    left2 = gSpriteX[sprite_index] >> X_FRACTION_BITS;
    image2_id = gSpriteImageID[sprite_index];
    if (left2 < right1)
    {
        double2 = (image2_id <= DOUBLE_SPRITES) ? 8 : 0;
        right2 = left2 + 8 + double2;
        if (right2 > left1)
        {
            top2 = gSpriteY[sprite_index] >> Y_FRACTION_BITS;
            if (top2 < bottom1)
            {
                bottom2 = top2 + gImageHeight[image2_id];
                if (bottom2 > top1)
                {
                    // at this point we've determined that the sprite rectangles overlap
                    
                    if (top1 > top2)
                    {
                        top3 = top1;
                        image1 = (unsigned char *)flashdata + gImageGraphics[image1_id];
                        image2 = (unsigned char *)flashdata + gImageGraphics[image2_id] + top1 - top2;
                    }
                    else
                    {
                        top3 = top2;
                        image1 = (unsigned char *)flashdata + gImageGraphics[image1_id] + top2 - top1;
                        image2 = (unsigned char *)flashdata + gImageGraphics[image2_id];
                    }
                    
                    bottom3 = (bottom1 < bottom2) ? bottom1 : bottom2;
                    
                    x_shift = left2 - left1;
                    y_overlap = bottom3-top3;
                    
                    if (double2)
                    {
                        if (x_shift < 0)
                        {
                            for(j=0;j<y_overlap;j++)
                                if ((image1[j] << 8) & (DoubleBits(image2[j]) << -x_shift))
                                    return 1;   // collision!
                        }
                        else
                        {
                            x_shift += 8;
                            for(j=0;j<y_overlap;j++)
                                if ((image1[j] << x_shift) & DoubleBits(image2[j]))
                                    return 1;   // collision!
                        }
                    }
                    else
                    {
                        if (x_shift < 0)
                        {
                            for(j=0;j<y_overlap;j++)
                                if (image1[j] & (image2[j] << -x_shift))
                                    return 1;   // collision!
                        }
                        else
                        {
                            for(j=0;j<y_overlap;j++)
                                if ((image1[j] << x_shift) & image2[j])
                                    return 1;   // collision!
                        }
                    }
                }
            }
        }
    }
    return 0;
}

const unsigned char quad_pair[] = {0, 0x0f, 0xf0, 0xff};

unsigned int QuadBits(int bits, int shift)
{
	// must cast as unsigned int so sign isn't maintained if there's a shift
	// ie: if bit 31 is on, we want it off after the shift
	return (unsigned int)((quad_pair[bits & 3]) +
                          (quad_pair[(bits >> 2) & 3] << 8) +
                          (quad_pair[(bits >> 4) & 3] << 16) +
                          (quad_pair[(bits >> 6) & 3] << 24)) >> shift;
}


int CollisionQuad(int image1_id, int left1, int top1, int sprite_index)
{
    int right1;
    int bottom1;
    
    int left2;
    int top2;
    int right2;
    int bottom2;
    int image2_id;
    
    int top3;
    int bottom3;
    
    int j;
    
    unsigned char * image1;
    unsigned char * image2;
    unsigned char * shift;
    
    int y_overlap;
    int x_shift;
    
    right1 = left1 + 8;
    bottom1 = top1 + gImageHeight[image1_id];
    
    left2 = gSpriteX[sprite_index] >> X_FRACTION_BITS;
    image2_id = gSpriteImageID[sprite_index];
    if (left2 < right1)
    {
        right2 = left2 + 32;
        if (right2 > left1)
        {
            top2 = gSpriteY[sprite_index] >> Y_FRACTION_BITS;
            if (top2 < bottom1)
            {
                bottom2 = top2 + gImageHeight[image2_id];
                if (bottom2 > top1)
                {
                    // at this point we've determined that the sprite rectangles overlap
                    
                    if (top1 > top2)
                    {
                        top3 = top1;
                        image1 = (unsigned char *)flashdata + gImageGraphics[image1_id];
                        image2 = (unsigned char *)(unsigned char *)queue + STATION_IMAGES + sprite_index*STATION_HEIGHT + top1 - top2;
                        shift = (unsigned char *)flashdata + gImageShift[image2_id] + top1 - top2;
                    }
                    else
                    {
                        top3 = top2;
                        image1 = (unsigned char *)flashdata + gImageGraphics[image1_id] + top2 - top1;
                        image2 = (unsigned char *)queue + STATION_IMAGES + sprite_index*STATION_HEIGHT;
                        shift = (unsigned char *)flashdata + gImageShift[image2_id];
                    }
                    
                    bottom3 = (bottom1 < bottom2) ? bottom1 : bottom2;
                    
                    x_shift = left2 - left1;
                    y_overlap = bottom3-top3;
                    
                    if (x_shift < 0)
                    {
                        for(j=0;j<y_overlap;j++)
                            if ((image1[j] << 24) & (QuadBits(image2[j], shift[j] & 7) << -x_shift))
                                return 1;   // collision!
                    }
                    else
                    {
                        x_shift += 24;
                        for(j=0;j<y_overlap;j++)
                            if ((image1[j] << x_shift) & QuadBits(image2[j], shift[j] & 7))
                                return 1;   // collision!
                    }
                }
            }
        }
    }
    return 0;
}

void ShipCollisions()
{
    int j;
    int x;
    int y;
    int start;
    int end;
    int sprite;
    
    if ((gSpriteState[PLAYER_SPRITE] & 0xC0) == 0x80)   // make sure player is not exploding
    {
        x = gSpriteX[PLAYER_SPRITE] >> X_FRACTION_BITS;
        y = gSpriteY[PLAYER_SPRITE] >> Y_FRACTION_BITS;
        start = ((y / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE;
        end = (((y + gImageHeight[gSpriteImageID[PLAYER_SPRITE]]) / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE + STATIONARY_DENSITY;
                                                 
                                                 
                                                 
        // check for player - station collision
        for (j=0; j<8; j++)
        {
            if (gSpriteState[j] & SS_STATION_ONSCREEN)  //
            {
                // not yet written
            }
        }
        
        // check for player - enemy ship collision
        for(j=FORMATION_SPRITE;j<FIRST_STATIONARY_SPRITE;j++)
        {
            if (gSpriteState[j]) // if ((gSpriteState[j] & 0xC0) == 0x80)
            {
                if (Collision(gSpriteImageID[PLAYER_SPRITE], x, y, j))
                {
                    InitDeathSequence(PLAYER_SPRITE, 0);
                    TriggerSoundEffect(SOUND_PLAYER_HIT_ID);
                }
            }
        }
        
        // check for player - stationary object collision
        for (j=start; j<end; j++)
        {
            if (j >= MAX_SPRITES)
                sprite = j - (32 * STATIONARY_DENSITY);     // this occurs if shot is straddling the vertical screen wrap point
            else
                sprite = j;
            if (gSpriteState[sprite] && (gSpriteState[sprite] & SS_EXPLOSION_NOT_DEADLY) == 0) // ((gSpriteState[sprite] & 0xC0) == 0x80)
            {
                if (Collision(gSpriteImageID[PLAYER_SPRITE], x, y, sprite))
                {
                    InitDeathSequence(PLAYER_SPRITE,0);
                    TriggerSoundEffect(SOUND_PLAYER_HIT_ID);

                    switch (gSpriteImageID[sprite])
                    {
                        case MINE_ID:
                            IncreaseScore(0x20);
                            InitDeathSequence(sprite, 0);
                            TriggerSoundEffect(SOUND_MINE_ID);
                            break;
                            
                        case ASTEROID_ID:
                        case ASTEROID_ID+1:
                        case ASTEROID_ID+2:
                            IncreaseScore(0x10);
                            InitDeathSequence(sprite, SS_EXPLOSION_NOT_DEADLY);
                            TriggerSoundEffect(SOUND_ASTEROID_ID);
                            break;
                    }
                }
            }
        }
    }
}

void ShotCollisions()
{
    int i;
    int j;
    int k;
    int x;
    int y;
    int start;
    int end;
    int sprite;
    int shot_id;
    int pod;
    
#ifdef QUAD_TEST
    i = MAX_SHOTS-1;
    x = (gShotX[i] + (gWindowX >> X_FRACTION_BITS)) & (WORLD_WIDTH-1);
    y = (gShotY[i] + (gWindowY >> Y_FRACTION_BITS)) & (WORLD_HEIGHT-1);
    shot_id = (gShotState[i] & 3) + SHOT_HORIZONTAL_ID;
    
    gScore = 0;
    
    // check for player shot-station collision
    for (j=0; j<8; j++)
    {
        if (gSpriteState[j] & SS_STATION_ONSCREEN)
        {
            pod = gSpriteImageID[j] * 7;  // horizontal = 0, vertical = 7
            if (gShotState[i])
                if (CollisionQuad(shot_id, x, y, j))
                    gScore = 0x888888;
        }
    }
    return;
#endif
    // process player shots
    for (i=0;i<4;i++)
    {
        if (gShotState[i])
        {
            x = (gShotX[i] + (gWindowX >> X_FRACTION_BITS)) & (WORLD_WIDTH-1);
            y = (gShotY[i] + (gWindowY >> Y_FRACTION_BITS)) & (WORLD_HEIGHT-1);
            shot_id =(gShotState[i] & 3) + SHOT_HORIZONTAL_ID;
            start = ((y / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE;
            end = (((y + shot_height[gShotState[i] & 3]) / 32) * STATIONARY_DENSITY) + FIRST_STATIONARY_SPRITE + STATIONARY_DENSITY;
            
            // check for player shot-station collision
            for (j=0; j<8; j++)
            {
                if (gSpriteState[j] & SS_STATION_ONSCREEN)
                {
                    pod = gSpriteImageID[j] * 7;  // horizontal = 0, vertical = 7
                    for(k=0;k<7;k++)
                    {
                        if (gStationState[j] & (1 << k)) // if pod is active
                        {
                            gSpriteX[MAX_SPRITES]      = gSpriteX[j] + (pod_offset_x[pod+k] << X_FRACTION_BITS);
                            gSpriteY[MAX_SPRITES]      = gSpriteY[j] + (pod_offset_y[pod+k] << Y_FRACTION_BITS);
                            gSpriteImageID[MAX_SPRITES]= pod_shape[pod+k];
                            if (Collision(shot_id, x, y, MAX_SPRITES))
                            {
                                if (k < 6)  // hit a pod
                                {
                                    gStationState[j] ^= (1 << k); // turn off pod
                                    IncreaseScore(0x50);
                                    TriggerSoundEffect(SOUND_MINE_ID);
                                    gShotState[i] = 0;
                                    
                                }
                                // destroy station if energy field  hit or last pod was hit
                                if (k == 6 || (gStationState[j] & 0x3f) == 0)
                                {
                                    gSpriteState[j] = 0;    // need to do something different...
                                    IncreaseScore(0x200);
                                    TriggerSoundEffect(SOUND_MINE_ID);
                                    gShotState[i] = 0;
                                }
                            }
                        }
                    }
                    if (gShotState[i])
                        if (CollisionQuad(shot_id, x, y, j))
                            gShotState[i] = 0;
                }
            }
            
            // check for player shot - enemy ship collision
            for(j=FORMATION_SPRITE;j<FIRST_STATIONARY_SPRITE;j++)
            {
                if (gSpriteState[PLAYER_SPRITE])    // CHANGE TO MAKE SURE SPRITE IS ACTIVE AND *NOT* EXPLODING
                {
                    // not yet written
                }
            }
            
            // check for player shot - stationary object collision
            for (j=start; j<end; j++)
            {
                if (j >= MAX_SPRITES)
                    sprite = j - (32 * STATIONARY_DENSITY);     // this occurs if shot is straddling the vertical screen wrap point
                else
                    sprite = j;
                if (gSpriteState[sprite])
                {
                    if (Collision(shot_id, x, y, sprite))
                    {
                        switch (gSpriteImageID[sprite])
                        {
                            case MINE_ID:
                                IncreaseScore(0x20);
                                InitDeathSequence(sprite, 0);
                                TriggerSoundEffect(SOUND_MINE_ID);
                                gShotState[i] = 0;
                                break;
                                
                            case ASTEROID_ID:
                            case ASTEROID_ID+1:
                            case ASTEROID_ID+2:
                                IncreaseScore(0x10);
                                InitDeathSequence(sprite, SS_EXPLOSION_NOT_DEADLY);
                                TriggerSoundEffect(SOUND_ASTEROID_ID);
                                gShotState[i] = 0;
                                break;
                        }
                    }
                }
            }
        }
    }
    
    // process station shots
#ifdef QUAD_TEST
    for (i=4;i<MAX_SHOTS - QUAD_TEST;i++)
#else
    for (i=4;i<MAX_SHOTS;i++)
#endif
    {
        if (gShotState[i])
        {
            if ((gSpriteState[PLAYER_SPRITE] & 0xC0) == 0x80)    // IF PLAYER IS ACTIVE AND *NOT* EXPLODING
            {
                x = (gShotX[i] + (gWindowX >> X_FRACTION_BITS)) & (WORLD_WIDTH-1);
                y = (gShotY[i] + (gWindowY >> Y_FRACTION_BITS)) & (WORLD_HEIGHT-1);

                // check for station shot - player collision
                if (Collision(SHOT_STATION_ID, x, y, PLAYER_SPRITE))
                {
                    InitDeathSequence(PLAYER_SPRITE, 0);
                    TriggerSoundEffect(SOUND_PLAYER_HIT_ID);
                    gShotState[i] = 0;
                }
            }
        }
    }
}


int BCDadd(register int a, register int b)
{
	// http://www.cs.uiowa.edu/~jones/bcd/bcd.html
	unsigned int t1, t2, t3, t4, t5, t6;
    
	t1 = a + 0x06666666;
	t2 = t1 + b;
	t3 = t1 ^ b;
	t4 = t2 ^ t3;
	t5 = ~t4 & 0x11111110;
	t6 = (t5 >> 2) | (t5 >> 3);
	return t2 - t6;
}

void IncreaseScore(int points)
{
    int temp;
//    int test1;
//    int test2;
    temp = BCDadd(gScore, points);
//    test1 = temp & 0xf000;
//    test2 = gScore & 0xf000;
//    
//    // bonus lives 0 = n/a 1=1K 2=2K 3=5K
//    if ((BONUS_LIFE) && (test1 != test2))
//    {
//        if ((BONUS_LIFE == 1) ||
//            (BONUS_LIFE == 2 && (test1 & 0x1000) == 0) ||
//            (BONUS_LIFE == 3 && (test1 == 0 || test1 == 0x5000)))
//        {
//            if (LIVES < 255) LIVES++;
//            TriggerSoundEffect(SOUND_EXTRA_HUMANOID_ID);
//        }
//    }
    gScore = temp;
}

void PlayerDecreaseLives()
{
    int i;
    int l;
    const unsigned short int *level;

    if (gReserveLives == 0)
    {
        //ActivateMainMenu();
        MODE = 0;
    }
    else
    {
        gReserveLives--;
        
        level = levels[gLevel];
        
        // position the player at start of level
        l = 0;
        gSpriteState[PLAYER_SPRITE]     = 0x80;
        gSpriteImageID[PLAYER_SPRITE]   = SHIP_ID;
        gSpriteColor[PLAYER_SPRITE]     = gObjectColors[COLOR_SHIP];
        gSpriteX[PLAYER_SPRITE]         = level[l++] << X_FRACTION_BITS; //80;
        gSpriteY[PLAYER_SPRITE]         = level[l++] << Y_FRACTION_BITS; //(DATASTREAM_SIZE - SHIP_HEIGHT) /2;
        gShipDir                        = 0;
        
        gWindowX = gSpriteX[PLAYER_SPRITE] - (80 << X_FRACTION_BITS);
        gWindowY = gSpriteY[PLAYER_SPRITE] - (((DATASTREAM_SIZE - SHIP_HEIGHT) /2) << Y_FRACTION_BITS);
        
        for (i=0;i<MAX_SHOTS;i++)
            gShotState[i] = 0;
    }
}


void InitDeathSequence(int index, int not_deadly)
{
    int adjust;

    // explosion height is 21 pixels
    adjust =  (21 - gImageHeight[gSpriteImageID[index]]) >> 1;
    if (not_deadly)
    {
        gSpriteState[index] = 0xC0 + (EXPLOSION_TIME >> 1) + SS_EXPLOSION_NOT_DEADLY;
        gSpriteColor[index] = 12;  // white if not deadly
    }
    else
    {
        gSpriteState[index] = 0xC0 + EXPLOSION_TIME;
        gSpriteColor[index] = FRAME;
    }
    gSpriteImageID[index] = EXPLOSION_ID;
    gSpriteX[index] -= 4 << X_FRACTION_BITS;
    gSpriteY[index] -= adjust << Y_FRACTION_BITS;
}

void DeathSequences()
{
    int i;
    int time;
    for (i=PLAYER_SPRITE; i< MAX_SPRITES;i++)
    {
        
        if ((gSpriteState[i] & 0xC0) == 0xC0)
        {
            if (gSpriteState[i] & SS_EXPLOSION_NOT_DEADLY)
            {
                time = (EXPLOSION_TIME >> 1); // faster explosion
            }
            else
            {
                time = EXPLOSION_TIME;
                gSpriteColor[i] = FRAME;
            }
            if (gSpriteState[i] & SS_EXPLOSION_TIME_MASK)
            {
                gSpriteState[i]--;
            }
            else
            {
                if (gSpriteImageID[i] < LAST_EXPLOSION_ID)
                {
                    gSpriteImageID[i]++;
                    gSpriteState[i] += time;
                }
                else
                {
                    gSpriteState[i] = 0;
                    if (i == PLAYER_SPRITE)
                        PlayerDecreaseLives();
                }
            }
        }
    }
}